#!/usr/bin/env python3
# coding: utf-8
from __future__ import print_function
from .__version__ import S_VERSION
from .__init__ import EP, WINDOWS, COLORS, unicode
from . import util as Util
from . import inetcat as Inetcat
from . import itelnet as Itelnet
from . import world as World

import os
import sys
import time
import signal
import select
import threading
from datetime import datetime

print = Util.print


"""r0c.py: retr0chat Telnet/Netcat Server"""
__author__ = "ed <a@ocv.me>"
__credits__ = ["stackoverflow.com"]
__license__ = "MIT"
__copyright__ = 2018
__url__ = "https://github.com/9001/r0c"


if "r0c" not in sys.modules:
    print(
        "\r\n  retr0chat must be launched as a module.\r\n  in the project root, run this:\r\n\r\n    python3 -m r0c\r\n"
    )
    sys.exit(1)


def optgen(ap, pwd):
    ac = ap
    u = unicode
    pt, pn = [23, 531] if WINDOWS or not os.getuid() else [2323, 1531]

    # fmt: off
    ac.add_argument("-i", metavar="IP", type=u, default="0.0.0.0", help="address to listen on")
    ac.add_argument("-pt", metavar="PORT", type=int, default=pt, help="telnet port (disable with 0)")
    ac.add_argument("-pn", metavar="PORT", type=int, default=pn, help="netcat port (disable with 0)")
    ac.add_argument("-tpt", metavar="PORT", type=int, default=0, help="TLS telnet port, e.g. 2424 (disable with 0)")
    ac.add_argument("-tpn", metavar="PORT", type=int, default=0, help="TLS netcat port, e.g. 1515 (disable with 0)")
    ac.add_argument("-pw", metavar="PWD", type=u, default=pwd, help="admin password")
    ac.add_argument("--nsalt", metavar="TXT", type=u, default="lammo/", help="salt for generated nicknames based on IP")

    ac = ap.add_argument_group("logging")
    ac.add_argument("--log-rx", action="store_true", help="log incoming traffic from clients")
    ac.add_argument("--log-tx", action="store_true", help="log outgoing traffic to clients")
    ac.add_argument("--rot-msg", metavar="N", type=int, default=131072, help="max num msgs per logfile")

    ac = ap.add_argument_group("tls")
    ac.add_argument("--ciphers", metavar="S", type=u, default="", help="specify allowed TLS ciphers; python default if unset")
    ac.add_argument("--tls-min", metavar="S", type=u, default="", help="oldest ver to allow; SSLv3 TLSv1 TLSv1_1 TLSv1_2 TLSv1_3")
    ac.add_argument("--old-tls", action="store_true", help="support old clients (centos6/powershell), bad ciphers")

    ac = ap.add_argument_group("perf")
    ac.add_argument("--hist-rd", metavar="N", type=int, default=65535, help="max num msgs to load from disk when joining a channel")
    ac.add_argument("--hist-mem", metavar="N", type=int, default=98303, help="max num msgs to keep in channel scrollback")
    ac.add_argument("--hist-tsz", metavar="N", type=int, default=16384, help="num msgs to discard when chat exceeds hist-mem")

    ac = ap.add_argument_group("debug")
    ac.add_argument("--dbg", action="store_true", help="show negotiations etc")
    ac.add_argument("--hex-rx", action="store_true", help="print incoming traffic from clients")
    ac.add_argument("--hex-tx", action="store_true", help="print outgoing traffic to clients")
    ac.add_argument("--hex-lim", metavar="N", type=int, default=128, help="filter packets larger than N bytes from being hexdumped")
    ac.add_argument("--hex-w", metavar="N", type=int, default=16, help="width of the hexdump, in bytes per line, mod-8")
    ac.add_argument("--dev", action="store_true", help="enable dangerous shortcuts (devmode)")
    ac.add_argument("--thr-mon", action="store_true", help="start monitoring threads on ctrl-c")
    if WINDOWS:
        ac.add_argument("--reuseaddr", action="store_true", help="allow rapid server restart (DANGER: lets you accidentally start multiple instances)")
    ac.add_argument("--linemode", action="store_true", help="force clients into linemode (to debug linemode UI)")
    ac.add_argument("--bench", action="store_true", help="dump statistics every 2 sec")
    # fmt: on


class Fargparse(object):
    def __init__(self):
        pass

    def add_argument_group(self, *a, **ka):
        return self

    def add_argument(self, opt, default=False, **ka):
        setattr(self, opt.lstrip("-").replace("-", "_"), default)


def run_fap(argv, pwd):
    ap = Fargparse()
    optgen(ap, pwd)

    if u"-h" in unicode(([""] + argv)[-1]):
        print()
        print("arg 1: Telnet port (0=disable), default: {0}".format(ap.pt))
        print("arg 2: NetCat port (0=disable), default: {0}".format(ap.pn))
        print("arg 3: admin password, default: {0}".format(ap.pw))
        print()
        sys.exit(0)

    try:
        setattr(ap, "pt", int(argv[1]))
        setattr(ap, "pn", int(argv[2]))
        setattr(ap, "pw", unicode(argv[3]))
    except:
        pass

    return ap


try:
    import argparse

    class RiceFormatter(argparse.HelpFormatter):
        def _get_help_string(self, action):
            """
            same as ArgumentDefaultsHelpFormatter(HelpFormatter)
            except the help += [...] line now has colors
            """
            fmt = "\033[36m (default: \033[35m%(default)s\033[36m)\033[0m"
            if not COLORS:
                fmt = " (default: %(default)s)"

            help = action.help
            if "%(default)" not in action.help:
                if action.default is not argparse.SUPPRESS:
                    defaulting_nargs = [argparse.OPTIONAL, argparse.ZERO_OR_MORE]
                    if action.option_strings or action.nargs in defaulting_nargs:
                        help += fmt
            return help

        def _fill_text(self, text, width, indent):
            """same as RawDescriptionHelpFormatter(HelpFormatter)"""
            return "".join(indent + line + "\n" for line in text.splitlines())

    class Dodge11874(RiceFormatter):
        def __init__(self, *args, **kwargs):
            kwargs["width"] = 9003
            super(Dodge11874, self).__init__(*args, **kwargs)

    def run_ap(argv, pwd):
        throw = False
        for formatter in [RiceFormatter, Dodge11874]:
            try:
                ap = argparse.ArgumentParser(formatter_class=formatter, prog="r0c")

                optgen(ap, pwd)
                return ap.parse_args(args=argv[1:])
            except AssertionError:
                if throw:
                    raise
                throw = True

except:
    run_ap = run_fap


class Core(object):
    def __init__(self):
        pass

    def start(self, argv=None):
        if WINDOWS and COLORS:
            os.system("rem")  # best girl

        if argv is None:
            argv = sys.argv

        for d in ["pm", "chan", "wire"]:
            try:
                os.makedirs(EP.log + d)
            except:
                pass

        print("  *  r0c {0}, py {1}".format(S_VERSION, Util.host_os()))

        pwd = "hunter2"
        pwd_file = os.path.join(EP.app, "password.txt")
        if os.path.isfile(pwd_file):
            print("  *  Password from " + pwd_file)
            with open(pwd_file, "rb") as f:
                pwd = f.read().decode("utf-8").strip()

        # old argv syntax compat
        try:
            _ = int(argv[1])
            rap = run_fap
        except:
            rap = run_ap

        ar = self.ar = rap(argv, pwd)  # type: argparse.Namespace
        Util.HEX_WIDTH = ar.hex_w
        Itelnet.init(ar)

        cert = EP.app + "cert.pem"
        if ar.tpt or ar.tpn:
            print("  *  Loading certificate {0}".format(cert))
            import ssl

            if not os.path.exists(cert):
                Util.builtins.print(
                    """\033[1;31m
tls was requested, but certificate not found at {0}pem
create the certificate (replacing "r0c.int" with the server's external ip or fqdn) and try again:

printf '%s\\n' GK . . . . r0c.int . | openssl req -newkey rsa:2048 -sha256 -keyout "{0}key" -nodes -x509 -days 365 -out "{0}crt" && cat "{0}key" "{0}crt" > "{0}pem"
\033[0m""".format(
                        EP.app + "cert."
                    )
                )
                raise Exception("TLS certificate not found")

            if ar.old_tls:
                ar.tls_min = ar.tls_min or "TLSv1"
                ar.ciphers = "EECDH+AESGCM:EDH+AESGCM:AES256+EECDH:AES256+EDH:ECDHE-RSA-AES128-GCM-SHA384:ECDHE-RSA-AES128-GCM-SHA256:ECDHE-RSA-AES128-GCM-SHA128:DHE-RSA-AES128-GCM-SHA384:DHE-RSA-AES128-GCM-SHA256:DHE-RSA-AES128-GCM-SHA128:ECDHE-RSA-AES128-SHA384:ECDHE-RSA-AES128-SHA128:ECDHE-RSA-AES128-SHA:ECDHE-RSA-AES128-SHA:DHE-RSA-AES128-SHA128:DHE-RSA-AES128-SHA128:DHE-RSA-AES128-SHA:DHE-RSA-AES128-SHA:ECDHE-RSA-DES-CBC3-SHA:EDH-RSA-DES-CBC3-SHA:AES128-GCM-SHA384:AES128-GCM-SHA128:AES128-SHA128:AES128-SHA128:AES128-SHA:AES128-SHA:DES-CBC3-SHA:HIGH:!aNULL:!eNULL:!EXPORT:!DES:!MD5:!PSK:!RC4"

        for srv, port in [
            ["Telnet", ar.pt],
            ["NetCat", ar.pn],
            ["TLS-Telnet", ar.tpt],
            ["TLS-NetCat", ar.tpn],
        ]:
            if port:
                print("  *  {0} server on port {1}".format(srv, port))
            else:
                print("  *  {0} server disabled".format(srv))

        if ar.pw == "hunter2":
            print("\033[1;31m")
            print("  using default password '{0}'".format(ar.pw))
            print("  change it with argument -pw or save it here:")
            print("  " + pwd_file)
            print("\033[0m")

        print("  *  Logs at " + EP.log)

        self.stopping = 0
        self.threadmon = False
        self.shutdown_flag = threading.Event()
        Util.py26_threading_event_wait(self.shutdown_flag)

        print("  *  Capturing ^C")
        for sig in [signal.SIGINT, signal.SIGTERM]:
            signal.signal(sig, self.signal_handler)

        print("  *  Creating world")
        self.world = World.World(self)

        self.servers = []
        for name, ctor, p1, p2, tls in [
            ["Telnet", Itelnet.TelnetServer, ar.pt, ar.pn, False],
            ["NetCat", Inetcat.NetcatServer, ar.pn, ar.pt, False],
            ["TLS-Telnet", Itelnet.TelnetServer, ar.tpt, ar.tpn, True],
            ["TLS-NetCat", Inetcat.NetcatServer, ar.tpn, ar.tpt, True],
        ]:
            if not p1:
                continue

            print("  *  Starting {0} server".format(name))
            srv = ctor(ar.i, p1, self.world, p2, tls)
            self.servers.append(srv)

        print("  *  Loading user configs")
        for server in self.servers:
            server.load_configs()

        print("  *  Starting push driver")
        self.push_thr = threading.Thread(
            target=self.push_worker,
            args=(self.world, self.servers),
            name="push",
        )
        # self.push_thr.daemon = True
        self.push_thr.start()

        print("  *  Running")
        self.select_thr = threading.Thread(target=self.select_worker, name="selector")
        self.select_thr.daemon = True
        self.select_thr.start()

        return True

    def run(self):
        print("  *  r0c is up  ^^,")

        if not self.ar.bench:
            try:
                timeout = 69
                if WINDOWS:
                    # ctrl-c does not raise
                    timeout = 0.69

                while not self.shutdown_flag.wait(timeout):
                    pass
            except:
                pass
        else:
            last_joins = 0
            last_parts = 0
            last_messages = 0
            while not self.shutdown_flag.is_set():
                for n in range(20):
                    if self.shutdown_flag.is_set():
                        break
                    time.sleep(0.1)

                print(
                    "{0:.3f}  j {1}  p {2}  m {3}  d {4},{5},{6}".format(
                        time.time(),
                        self.world.num_joins,
                        self.world.num_parts,
                        self.world.num_messages,
                        self.world.num_joins - last_joins,
                        self.world.num_parts - last_parts,
                        self.world.num_messages - last_messages,
                    )
                )

                last_joins = self.world.num_joins
                last_parts = self.world.num_parts
                last_messages = self.world.num_messages

        # termiante refresh_chans
        self.world.dirty_ch = {}
        self.world.dirty_flag.set()

        with self.world.mutex:
            pass

        print("  *  saving user configs")
        for server in self.servers:
            server.save_configs()

        print("  *  terminating world")
        self.world.shutdown()

        print("  *  selector cleanup")
        for server in self.servers:
            server.srv_sck.close()

        print("  *  r0c is down")
        return True

    def select_worker(self):
        srvs = {}
        for iface in self.servers:
            srvs[iface.srv_sck] = iface

        sn = -1
        sc = {}
        slow = {}  # sck:cli
        fast = {}
        nfast = 0
        dirty_ref = 0
        next_slow = 0
        timeout = None
        while not self.shutdown_flag.is_set():
            nsn = self.world.cserial
            if sn != nsn:
                sn = nsn
                sc = {}
                slow = {}
                fast = {}
                for srv in self.servers:
                    for c in srv.clients:
                        if c.slowmo_tx or (c.wizard_stage is not None and not c.is_bot):
                            slow[c.socket] = c
                        else:
                            fast[c.socket] = c

                        sc[c.socket] = c

                timeout = 0.2 if slow else 1 if fast else 69

            want_tx = [s for s, c in fast.items() if c.writable()]
            want_rx = [s for s, c in sc.items() if c.readable()]
            want_rx += list(srvs.keys())

            now = time.time()
            if slow and now >= next_slow:
                next_slow = now + 0.18

                for c in slow.values():
                    if c.slowmo_skips:
                        c.slowmo_skips -= 1

                want_tx += [
                    s for s, c in slow.items() if c.writable() and not c.slowmo_skips
                ]

            if dirty_ref != self.world.last_dirty or self.world.dirty_flag.is_set():
                dirty_ref = self.world.last_dirty
                nfast = 0
            else:
                nfast += 1

            ct = 0.09 if nfast < 2 else timeout

            try:
                # print("sel", len(want_rx), len(want_tx), ct)
                rxs, txs, _ = select.select(want_rx, want_tx, [], ct)
                if self.stopping:
                    break

                with self.world.mutex:
                    if sn != self.world.cserial:
                        continue

                    for s in rxs:
                        if s in srvs:
                            srvs[s].handle_accept()
                        else:
                            sc[s].handle_read()

                    for s in txs:
                        sc[s].handle_write()

            except Exception as ex:
                if "Bad file descriptor" in str(ex):
                    # print('osx bug ignored')
                    continue
                Util.whoops()

    def push_worker(self, world, ifaces):
        last_action_ts = time.time()
        last_its = None
        last_date = None
        c_refresh = 0
        while not self.shutdown_flag.is_set():
            if any(srv.clients for srv in self.servers):
                # sleep until the start of the next mod5 utc second
                while True:
                    ts = time.time()
                    its = int(ts / 5) * 5
                    if its != last_its:
                        last_its = its
                        break
                    if ts - its < 4.99:
                        if self.shutdown_flag.wait((5 - (ts - its))):
                            break
                    else:
                        time.sleep(0.02)
            else:
                # less precision if there's nobody connected
                self.world.dirty_flag.wait(100)
                ts = time.time()
                last_its = int(ts / 5) * 5

            if c_refresh < 6:
                c_refresh += 1
            else:
                c_refresh = 1

            with world.mutex:
                if self.stopping:
                    break

                zd = datetime.utcfromtimestamp(ts)
                date = u"%04d-%02d-%02d" % (zd.year, zd.month, zd.day)

                if date != last_date:
                    if last_date:
                        world.broadcast_message(
                            u"\033[36mday changed to \033[1m{0}".format(date), False
                        )
                    last_date = date

                for iface in ifaces:
                    for client in iface.clients:
                        if client.handshake_sz and not c_refresh % client.m_refresh:
                            client.refresh(False)

                    # print('next scheduled kick: {0}'.format('x' if iface.next_scheduled_kick is None else iface.next_scheduled_kick - ts))

                    if (
                        iface.next_scheduled_kick is not None
                        and iface.next_scheduled_kick <= ts
                    ):
                        to_kick = []
                        next_min = None
                        for sch in iface.scheduled_kicks:
                            if sch[0] <= ts:
                                to_kick.append(sch)
                            else:
                                if next_min is None or next_min > sch[0]:
                                    next_min = sch[0]

                        for sch in to_kick:
                            timeout, remote, msg = sch
                            iface.scheduled_kicks.remove(sch)
                            if remote in iface.clients:
                                if msg is None:
                                    iface.part(remote)
                                else:
                                    iface.part(remote, False)
                                    print(msg)

                        iface.next_scheduled_kick = next_min

                if ts - last_action_ts >= 600:
                    last_action_ts = ts

                    # flush client configs
                    for iface in ifaces:
                        iface.save_configs()

                        # flush wire logs
                        if self.ar.log_rx or self.ar.log_tx:
                            for client in iface.clients:
                                if client.wire_log:
                                    try:
                                        client.wire_log.flush()
                                    except:
                                        Util.whoops()

                    # flush chan logs
                    for chan_list in [world.pub_ch, world.priv_ch]:
                        for chan in chan_list:
                            if chan.log_fh:
                                try:
                                    chan.log_fh.flush()
                                except:
                                    Util.whoops()

        print("  *  terminated push_worker")

    def shutdown(self):
        # monitor_threads()
        self.stopping += 1
        if self.stopping >= 3:
            os._exit(1)

        self.shutdown_flag.set()

    def signal_handler(self, sig, frame):
        if self.ar.thr_mon and not self.threadmon:
            self.threadmon = True
            Util.monitor_threads()
        else:
            self.shutdown()


def start_r0c(argv):
    core = Core()
    try:
        if core.start(argv):
            return core.run()
    except SystemExit:
        raise
    except:
        Util.whoops()
        os._exit(1)


def main(argv=None):
    mode = "normal"
    # mode = "profiler"
    # mode = 'test-ansi-annotation'
    # test_hexdump()

    if mode == "normal":
        if not start_r0c(argv):
            sys.exit(1)


"""
    if mode == "profiler":
        print("  *  PROFILER ENABLED")
        statfile = "profiler-results"
        import yappi

        yappi.start()
        start_r0c(argv)
        yappi.stop()

        fn_stats = yappi.get_func_stats()
        thr_stats = yappi.get_thread_stats()

        print()
        for ext in ["pstat", "callgrind", "ystat"]:
            print("writing {0}.{1}".format(statfile, ext))
            fn_stats.save("{0}.{1}".format(statfile, ext), type=ext)

        with open("{0}.func".format(statfile), "w") as f:
            fn_stats.print_all(out=f)

        with open("{0}.thr".format(statfile), "w") as f:
            thr_stats.print_all(out=f)

        print("\n\n{0}\n  func stats\n{0}\n".format("-" * 72))
        fn_stats.print_all()

        print("\n\n{0}\n  thread stats\n{0}\n".format("-" * 72))
        thr_stats.print_all()

    if mode == "test-ansi-annotation":
        Util.test_ansi_annotation()
"""


if __name__ == "__main__":
    main()



# coding: utf-8
from __future__ import print_function
from .__init__ import EP, PY2, INTERP, TYPE_CHECKING
from . import util as Util
from . import chat as Chat

import os
import re
import time
import threading
from datetime import datetime

if PY2:
    from Queue import Queue
else:
    from queue import Queue

if TYPE_CHECKING:
    from . import __main__ as Main
    from . import user as User

print = Util.print


class World(object):
    def __init__(self, core):
        # type: (Main.Core) -> World
        self.core = core
        self.ar = core.ar
        self.users = []  # User instances
        self.lusers = {}  # lowercase hashmap
        self.pub_ch = []  # NChannel instances (public)
        self.priv_ch = []  # NChannel instances (private)
        self.dirty_ch = {}  # Channels that have pending tx
        self.cntab = {}  # nick color cache
        self.cserial = 0  # selector configuration serial number
        self.task_queue = Queue()  # Delayed processing of expensive tasks
        self.mutex = threading.RLock()
        self.dirty_flag = threading.Event()  # raise after setting dirty_ch
        self.last_dirty = 0  # scheduler hint
        Util.py26_threading_event_wait(self.dirty_flag)

        # config
        self.messages_per_log_file = self.ar.rot_msg

        # stats for benchmarking
        self.num_joins = 0
        self.num_parts = 0
        self.num_messages = 0

        threading.Thread(target=self.refresh_chans, name="tx_chan").start()

    def shutdown(self):
        with self.mutex:
            for chanlist in [self.pub_ch, self.priv_ch]:
                ucs = [y for x in chanlist for y in x.uchans]
                ucs = [x for x in ucs if x.alias != u"r0c-status"]
                for uc in ucs:
                    self.part_chan(uc, True)

    def add_user(self, user):
        with self.mutex:
            self.users.append(user)
            if user.lnick:
                self.lusers[user.lnick] = user

    def find_user(self, nick):
        with self.mutex:
            return self.lusers.get(nick.lower(), None)

    def refresh_chans(self):
        while not self.core.shutdown_flag.is_set():
            # no latency unless multiple messages within 0.05 sec
            time.sleep(0.05)

            self.dirty_flag.wait()
            with self.mutex:
                self.dirty_flag.clear()
                if not self.dirty_ch:
                    # raised with no pending work; maybe shutdown signal
                    continue

                self.last_dirty = time.time()
                dirty_ch = list(self.dirty_ch)
                self.dirty_ch = {}

                for chan in dirty_ch:
                    self.refresh_chan(chan)

        print("  *  terminated refresh_chans")

    def refresh_chan(self, nchan):
        if not nchan.uchans:
            # all users left since this channel got scheduled for refresh
            return

        last_msg = nchan.msgs[-1].sno if nchan.msgs else 0

        upd_users = set()  # users with clients to repaint

        for uchan in nchan.uchans:
            if uchan.update_activity_flags(False, last_msg):
                # update status bar immediately
                if uchan.user.active_chan != uchan:
                    upd_users.add(uchan.user)

            if uchan.user.active_chan == uchan:
                if (
                    not uchan.user.client.handshake_sz
                    or uchan.user.client.wizard_stage is not None
                ):
                    continue

                # print('refreshing {0} for {1}'.format(nchan.get_name(), uchan.user.nick))
                upd_users.add(uchan.user)

        for user in upd_users:
            user.client.refresh(False)

    def send_chan_msg(self, from_nick, nchan, text, ping_self=True):
        max_hist_mem = self.ar.hist_mem
        msg_trunc_size = self.ar.hist_tsz
        with self.mutex:
            self.num_messages += 1
            if nchan.name is None and not from_nick.startswith(u"-"):
                # private chan, check if we have anyone to send to
                if len(nchan.uchans) == 1:
                    if nchan.uchans[0].alias == u"r0c-status":
                        if nchan.uchans[0].user.nick == from_nick:
                            self.send_chan_msg(
                                u"-err-",
                                nchan,
                                u"this buffer does not accept messages, only commands\n",
                            )
                            return

                    else:
                        # private chat without the other user added yet;
                        # pull in the other user
                        utarget = None
                        target = nchan.uchans[0].alias
                        if target != from_nick:
                            utarget = self.find_user(target)
                            if utarget is None:
                                self.send_chan_msg(
                                    u"-err-",
                                    nchan,
                                    u'\033[1;31mfailed to locate user "{0}"'.format(
                                        nchan.uchans[0].alias
                                    ),
                                )
                                return

                            self.join_chan_obj(utarget, nchan, from_nick)
                            self.start_logging(nchan)
                            # fallthrough to send message

            now = time.time()
            msg = Chat.Message(nchan, now, from_nick, text)
            nchan.msgs.append(msg)
            nchan.latest = msg.ts

            if not from_nick.startswith(u"-") and not from_nick == u"***":
                nchan.user_act_ts[from_nick] = now
                nchan.update_usernames()

            if len(nchan.msgs) > max_hist_mem:
                new_len = len(nchan.msgs) - msg_trunc_size
                print(
                    " hist trunc:  [{0}] from {1} to {2}".format(
                        nchan.get_name(), len(nchan.msgs), new_len
                    )
                )
                while new_len > max_hist_mem:
                    print("\033[1;31!!!\033[0m")
                    new_len -= msg_trunc_size
                nchan.msgs = nchan.msgs[msg_trunc_size:]

            # self.refresh_chan(nchan)
            for uchan in nchan.uchans:
                if nchan.name is None or uchan.user.nick_re.search(text):
                    # if len(nchan.uchans) == 1:
                    # 	break
                    if PY2 and INTERP != "IronPython":
                        if isinstance(uchan.user.nick, str):
                            Util.whoops("uchan.user.nick is bytestring")
                        if isinstance(from_nick, str):
                            Util.whoops("from_nick is bytestring")

                    if uchan.alias == u"r0c-status":
                        if ping_self:
                            uchan.last_ping = msg.sno
                    else:
                        if ping_self or uchan.user.nick != from_nick:
                            uchan.last_ping = msg.sno

            if nchan not in self.dirty_ch:
                self.dirty_ch[nchan] = 1
                self.dirty_flag.set()

            if nchan.log_fh:
                # print('logrotate counter at {0}'.format(nchan.log_ctr))
                if nchan.log_ctr >= self.messages_per_log_file:
                    self.start_logging(nchan)

                ltxt = u"%s %s %s\n" % (
                    hex(int(msg.ts * 8.0))[2:].rstrip("L"),
                    msg.user,
                    msg.txt,
                )
                nchan.log_fh.write(ltxt.encode("utf-8"))
                nchan.log_ctr += 1

    def join_chan_obj(self, user, nchan, alias=None):
        # type: (User.User, Chat.NChannel, str) -> Chat.UChannel
        with self.mutex:
            # print('{0} users in {1}, {2} messages; {3} is in {4} channels'.format(
            # 	len(nchan.uchans), nchan.get_name(), len(nchan.msgs), user.nick, len(user.chans)))

            for uchan in user.chans:
                if uchan.nchan == nchan:
                    return uchan

            self.num_joins += 1
            uchan = Chat.UChannel(user, nchan, alias)
            user.chans.append(uchan)
            nchan.uchans.append(uchan)
            nchan.user_act_ts[user.nick] = time.time()
            nchan.update_usernames()
            self.send_chan_msg(
                u"--",
                nchan,
                u"\033[1;32m{0}\033[22m has joined".format(user.nick),
                False,
            )
            if not alias:
                uchan.last_read = -1

            return uchan

    def get_pub_chan(self, name):
        for ch in self.pub_ch:
            if ch.name == name:
                return ch
        return None

    def get_priv_chan(self, user, alias):
        for ch in user.chans:
            if ch.alias == alias:
                return ch
        return None

    def join_pub_chan(self, user, name):
        with self.mutex:
            name = name.strip().lower()
            nchan = self.get_pub_chan(name)
            if nchan is None:
                nchan = Chat.NChannel(
                    name, u"#{0} - no topic has been set".format(name)
                )
                nchan.msgs.append(
                    Chat.Message(
                        nchan,
                        time.time(),
                        u"--",
                        u"\033[36mchannel created at \033[1m{0}".format(
                            datetime.utcnow().strftime("%Y-%m-%d, %H:%M:%SZ")
                        ),
                    )
                )
                if nchan.name != u"scrolltest":
                    self.load_chat_log(nchan)
                    # self.task_queue.put([self.load_chat_log, [nchan], {}])
                else:
                    for n1 in range(10):
                        txt = u""
                        for n2 in range(10):
                            txt += u"{0}{1}".format(n1, n2) * 20 + u" "

                        msg = Chat.Message(nchan, time.time(), u"--", txt)
                        nchan.msgs.append(msg)

                    txt = u"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa " * 6
                    msg = Chat.Message(nchan, time.time(), u"--", txt)
                    nchan.msgs.append(msg)

                self.pub_ch.append(nchan)

            ret = self.join_chan_obj(user, nchan)
            user.new_active_chan = ret
            return ret

    def join_priv_chan(self, user, alias):
        with self.mutex:
            uchan = self.get_priv_chan(user, alias)
            if uchan is None:
                nchan = Chat.NChannel(None, u"DM with [[uch_a]]")
                self.priv_ch.append(nchan)
                uchan = self.join_chan_obj(user, nchan)
                uchan.alias = alias
            return uchan

    """
    def broadcast_banner(self, msg):
        with self.mutex:
            chans = {}
            for user in self.users:
                if user.active_chan and user.active_chan.nchan not in chans:
                    chans[user.active_chan.nchan] = 1

            if not msg:
                for nchan in chans:
                    if nchan.topic_bak is not None:
                        nchan.topic = nchan.topic_bak
                        nchan.topic_bak = None
                for user in self.users:
                    if user.active_chan:
                        user.client.refresh(False)
            else:
                for nchan in chans:
                    if nchan.topic_bak is None:
                        nchan.topic_bak = nchan.topic
                    nchan.topic = msg

                # print('broadcast: {0}'.format(msg))
                for user in self.users:
                    if user.active_chan:
                        # print('         : {0} ->'.format(user))
                        to_send = u"\033[H{0}\033[K".format(msg)
                        user.client.screen[0] = to_send
                        user.client.say(
                            to_send.encode(user.client.codec, "backslashreplace")
                        )
    """

    def broadcast_message(self, msg, severity=1):
        """1=append, 2=append+scroll"""
        with self.mutex:
            msg = Util.convert_color_codes(msg, False)

            for nchan in self.pub_ch + self.priv_ch:
                self.send_chan_msg(u"-!-", nchan, msg)

            if severity > 1:
                for user in self.users:
                    if user.active_chan:
                        if not user.active_chan.lock_to_bottom:
                            user.active_chan.lock_to_bottom = True
                            user.client.need_full_redraw = True
                    else:
                        user.client.say(
                            u"\n [[ broadcast message ]]\n {0}\n".format(msg)
                            .replace(u"\n", u"\r\n")
                            .encode("utf-8")
                        )

    def part_chan(self, uchan, quiet=False):
        # type: (Chat.UChannel, bool) -> None
        with self.mutex:
            self.num_parts += 1
            user = uchan.user
            nchan = uchan.nchan
            i = None
            if user.active_chan == uchan:
                i = user.chans.index(uchan)
            user.chans.remove(uchan)
            nchan.uchans.remove(uchan)

            try:
                del nchan.user_act_ts[user.nick]
                nchan.update_usernames()
            except:
                pass

            if i:
                if len(user.chans) <= i:
                    i -= 1
                user.new_active_chan = user.chans[i]

            if not quiet:
                suf = u""
                if not nchan.name:
                    suf = u"; reinvite by typing another msg here"

                self.send_chan_msg(
                    u"--",
                    nchan,
                    u"\033[1;33m{0}\033[22m has left{1}".format(user.nick, suf),
                )

            if not nchan.uchans:
                print(" close chan:  [{0}]".format(nchan.get_name()))

                if nchan.log_fh:
                    nchan.log_fh.close()

                ch_list = self.pub_ch
                if not nchan.name:
                    ch_list = self.priv_ch

                ch_list.remove(nchan)

    def load_chat_log(self, nchan):
        if not nchan:
            return

        # print('  chan hist:  scanning files')
        t1 = time.time()

        log_dir = u"{0}chan/{1}".format(EP.log, nchan.name)
        try:
            os.makedirs(log_dir)
        except:
            pass

        if PY2:
            # os.walk stats all files (bad over nfs)
            files = os.listdir(log_dir.encode("utf-8"))
        else:
            files = []
            for (dirpath, dirnames, filenames) in os.walk(log_dir):
                files.extend(filenames)
                break

        re_chk = re.compile("^[0-9]{4}-[0-9]{4}-[0-9]{6}-*$")

        # total_size = 0
        # for fn in sorted(files):
        # 	total_size += os.path.getsize(
        # 		'{0}/{1}'.format(log_dir, fn))
        #
        # do_broadcast = (total_size > 1024*1024)
        # if do_broadcast:
        # 	self.broadcast_banner(u'\033[1;37;45m [ LOADING CHATLOG ] \033[0;42m')
        # 	# daily dose

        # print('  chan hist:  reading files')
        ln = u"???"
        t2 = time.time()
        chunks = [nchan.msgs]
        n_left = self.ar.hist_rd - len(nchan.msgs)
        bytes_loaded = 0
        try:
            for fn in sorted(files, reverse=True):
                if not re_chk.match(fn):
                    # unexpected file in log folder, skip it
                    continue

                chunk = []
                with open("{0}/{1}".format(log_dir, fn), "rb") as f:
                    f.readline()  # discard version info
                    for ln in f:
                        ts, user, txt = ln.decode("utf-8").rstrip(u"\n").split(u" ", 2)

                        chunk.append(Chat.Message(None, int(ts, 16) / 8.0, user, txt))

                    bytes_loaded += f.tell()

                # if chunk:
                # 	chunk.append(Message(None, int(ts, 16)/8.0, '--', \
                # 		'\033[36mend of log file "{0}"'.format(fn)))

                if len(chunk) > n_left:
                    chunk = chunk[-n_left:]

                chunks.append(chunk)
                n_left -= len(chunk)
                if n_left <= 0:
                    break
        except:
            Util.whoops(ln)

        # print('  chan hist:  merging {0} chunks'.format(len(chunks)))
        nchan.msgs = []
        for chunk in reversed(chunks):
            # print('\nadding {0} messages:\n  {1}'.format(
            # 	len(chunk), '\n  '.join(str(x) for x in chunk)))
            nchan.msgs.extend(chunk)

        # print('  chan hist:  setting {0} serials'.format(len(nchan.msgs)))
        for n, msg in enumerate(nchan.msgs):
            msg.sno = n

        t3 = time.time()
        print(
            "  chan hist:  {0} msgs, {1:.0f} kB, {2:.2f} + {3:.2f} sec, #{4}".format(
                self.ar.hist_rd - n_left,
                bytes_loaded / 1024.0,
                t2 - t1,
                t3 - t2,
                nchan.name,
            )
        )

        # if do_broadcast:
        # 	self.broadcast_banner(None)

        # if nchan.name != 'xst':
        self.start_logging(nchan, chunks[0])

    def start_logging(self, nchan, chat_backlog=None):
        if nchan.name is not None:
            log_dir = u"{0}chan/{1}".format(EP.log, Util.sanitize_fn(nchan.name))
        else:
            log_dir = u"{0}pm/{1}".format(
                EP.log, u"/".join([Util.sanitize_fn(x.user.nick) for x in nchan.uchans])
            )

        if nchan.log_fh:
            nchan.log_fh.close()

        # always mkdir because direction can turn on reconnect
        try:
            os.makedirs(log_dir)
        except:
            if not os.path.isdir(log_dir):
                raise

        ts = datetime.utcnow().strftime("%Y-%m%d-%H%M%S")
        log_fn = u"{0}/{1}".format(log_dir, ts)

        while os.path.isfile(log_fn):
            log_fn += u"-"

        nchan.log_ctr = 0
        nchan.log_fh = open(log_fn, "wb")
        nchan.log_fh.write(u"1 {0:x}\n".format(int(time.time())).encode("utf-8"))

        # print('opened log file {0}'.format(log_fn))

        if chat_backlog:
            # print('appending backlog ({0} messages)'.format(len(chat_backlog)))
            for msg in chat_backlog:
                ltxt = u"%s %s %s\n" % (
                    hex(int(msg.ts * 8.0))[2:].rstrip("L"),
                    msg.user,
                    msg.txt,
                )
                nchan.log_fh.write(ltxt.encode("utf-8"))
                nchan.log_ctr += 1

            # potential chance that a render goes through
            # before the async job processor kicks in
            self.dirty_ch[nchan] = 1
            self.dirty_flag.set()
            for uchan in nchan.uchans:
                uchan.user.client.need_full_redraw = True



# coding: utf-8
from __future__ import print_function
from . import util as Util
from . import chat as Chat

import sys
import gc

print = Util.print


def memory_dump():
    import _pickle as cPickle

    with open("profiler-results.memory", "wb") as dump:
        for obj in gc.get_objects():
            i = id(obj)
            size = sys.getsizeof(obj, 0)
            #    referrers = [id(o) for o in gc.get_referrers(obj) if hasattr(o, '__class__')]
            referents = [
                id(o) for o in gc.get_referents(obj) if hasattr(o, "__class__")
            ]
            if hasattr(obj, "__class__"):
                cls = str(obj.__class__)
                cPickle.dump(
                    {"id": i, "class": cls, "size": size, "referents": referents}, dump
                )


"""

def get_obj_name(target_id):
    variables = {}
    variables.extend(locals())
    variables.extend(globals())
    found = None
    var_id = None
    for var in variables:
        exec("var_id=id({0})".format(var))
        if var_id == target_id:
            exec("found={0}".format(var))
    if found is not None:
        print(found)
        print(id(found))
    else:
        print("/")


def find_leaked_messages():
    n_hits = 0
    for obj in gc.get_objects():
        if not type(obj) is Chat.Message:
            continue
        n_hits += 1
        if n_hits > 1000:
            break

        ref_objs = gc.get_referents(obj)
        referents = [id(o) for o in ref_objs if hasattr(o, "__class__")]
        print("obj,ref:", id(obj), referents)
        print("obj:", obj)
        print("ref:", ref_objs)
        print()

"""


# repl notepad
"""

from .diag import *
find_leaked_messages()

n_hits = 0
for obj in gc.get_objects():
  if not hasattr(obj, '__class__'):
    continue
  if type(obj) is Message:
    n_hits += 1
    if n_hits > 10000:
      break
    i = id(obj)
    size = sys.getsizeof(obj, 0)
    name = obj.__name__ if hasattr(obj, '__name__') else '/'
    referents = [id(o) for o in gc.get_referents(obj) if hasattr(o, '__class__')]
    print('id: {0}, size: {1}, class: {2}, name: {3}, referents: {4}'.format(
      i, size, str(obj.__class__), name, ', '.join([str(x) for x in referents])))

"""



# coding: utf-8
from __future__ import print_function
from .__init__ import EP, unicode
from . import util as Util
from . import ivt100 as Ivt100

import time

print = Util.print


class NetcatServer(Ivt100.VT100_Server):
    def __init__(self, host, port, world, other_if, tls):
        Ivt100.VT100_Server.__init__(self, host, port, world, other_if, tls)
        ucp = "{0}cfg.{1}netcat".format(EP.log, "tls-" if tls else "")
        self.user_config_path = ucp

    def gen_remote(self, socket, addr, user):
        return NetcatClient(self, socket, addr, self.world, user)


class NetcatClient(Ivt100.VT100_Client):
    def __init__(self, host, socket, address, world, user):
        Ivt100.VT100_Client.__init__(self, host, socket, address, world, user)

        self.looks_like_telnet = {
            b"\xff\xfe": 1,
            b"\xff\xfd": 1,
            b"\xff\xfc": 1,
            b"\xff\xfb": 1,
        }
        # trick telnet into revealing itself:
        # request client status and location
        self.replies.append(b"\xff\xfd\x05\xff\xfd\x17")

    def handle_read(self):
        with self.world.mutex:
            if self.dead:
                print("XXX reading when dead")
                return

            try:
                data = self.socket.recv(8192)
                if not data:
                    raise Exception()
            except:
                if not self.dead:
                    self.host.part(self)
                return

            if self.ar.hex_rx:
                Util.hexdump(data, "-->>")

            if self.wire_log and self.ar.log_rx:
                self.wire_log.write(
                    unicode(int(time.time() * 1000)).encode("utf-8") + b"\n"
                )
                Util.hexdump(data, ">", self.wire_log)

            self.in_bytes += data

            if b"\xff" in data:
                ofs = 0
                while ofs >= 0:
                    ofs = data.find(b"\xff", ofs)
                    if ofs < 0:
                        break
                    if data[ofs : ofs + 2] in self.looks_like_telnet:
                        self.num_telnet_negotiations += 1
                    ofs = ofs + 1
            try:
                src = unicode(self.in_bytes.decode(self.codec))
                self.in_bytes = self.in_bytes[0:0]

            except UnicodeDecodeError as uee:
                uee.start += self.uee_offset
                if len(self.in_bytes) < uee.start + 6:
                    print(
                        "need more data to parse unicode codepoint at {0} in {1}".format(
                            uee.start, len(self.in_bytes)
                        )
                    )
                    Util.hexdump(self.in_bytes[-8:], "XXX ")
                    src = u"{0}".format(self.in_bytes[: uee.start].decode(self.codec))
                    self.in_bytes = self.in_bytes[uee.start :]
                else:
                    # it can't be helped
                    print("warning: unparseable data:")
                    Util.hexdump(self.in_bytes, "XXX ")
                    src = u"{0}".format(
                        self.in_bytes[: uee.start].decode(
                            self.codec, "backslashreplace"
                        )
                    )
                    self.in_bytes = self.in_bytes[0:0]  # todo: is this correct?

            self.in_text += src

            if self.wizard_stage is not None and len(self.in_text_full) < 1024:
                self.in_text_full += src

            self.read_cb(False, len(src))



# coding: utf-8
from __future__ import print_function
from .__init__ import EP, PY2, WINDOWS, COLORS, INTERP, unicode

import traceback
import threading
import struct
import time
import sys
import os
import platform
import itertools
from datetime import datetime


print_mutex = threading.Lock()
if PY2:
    import __builtin__ as builtins
else:
    import builtins


HEX_WIDTH = 16

azAZ = u"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"

BRI_256 = list(
    itertools.chain(
        range(35, 47),
        range(48, 52),
        range(70, 88),
        range(103, 124),
        range(132, 232),
    )
)


def print(*args, **kwargs):
    args = list(args)
    try:
        if not COLORS and u"\033" in args[0]:
            args[0] = strip_ansi(args[0])
    except:
        pass

    with print_mutex:
        zd = datetime.utcnow()
        t = "%06d " % (
            (zd.hour * 100 + zd.minute) * 100 + zd.second,
            # zd.microsecond // 1000
        )
        builtins.print(
            t + str(args[0] if args else u"").replace(u"\n", u"\n" + t),
            *args[1:],
            **kwargs
        )


def num(c):
    try:
        return int(c)
    except:
        return None


def b2hex(data):
    if PY2:
        return " ".join(map(lambda b: format(ord(b), "02x"), data))
    else:
        if type(data) is str:
            return " ".join(map(lambda b: format(ord(b), "02x"), data))
        else:
            return " ".join(map(lambda b: format(b, "02x"), data))


def hexdump(pk, prefix="", file=None):
    if file is not None:
        line_fmt = u"{0} {2}{3}{4}"
        hex_width = 4
        blk_width = 4
    else:
        line_fmt = u"{0}{1:8x}  {2}{3} {4}"
        hex_width = HEX_WIDTH
        blk_width = 8

    lpk = len(pk)
    ofs = 0
    hexofs = 0
    hexlen = 0
    hexstr = ""
    ascstr = ""
    ascstr_width = int(hex_width * 100 / 32.0 - 0.5)  # 32h = 100a, 16h = 50a
    while ofs < lpk:
        hexstr += b2hex(pk[ofs : ofs + blk_width])
        hexstr += " "
        if PY2:
            ascstr += "".join(
                map(
                    lambda b: b if ord(b) >= 0x20 and ord(b) < 0x7F else ".",
                    pk[ofs : ofs + blk_width],
                )
            )
        else:
            ascstr += "".join(
                map(
                    lambda b: chr(b) if b >= 0x20 and b < 0x7F else ".",
                    pk[ofs : ofs + blk_width],
                )
            )
        hexlen += blk_width
        ofs += blk_width

        if hexlen >= hex_width or ofs >= lpk:
            txt = line_fmt.format(
                prefix, hexofs, hexstr, u" " * (ascstr_width - len(hexstr)), ascstr
            )

            if file is not None:
                file.write((txt + u"\n").encode("utf-8"))
            else:
                print(txt)

            hexofs = ofs
            hexstr = ""
            hexlen = 0
            ascstr = ""
        else:
            hexstr += " "
            ascstr += " "


"""

def test_hexdump():
    try:
        from StringIO import StringIO as bio
    except:
        from io import BytesIO as bio

    v = b""
    for n in range(5):
        print()
        v += b"a"
        fobj = bio()
        hexdump(v, ">", fobj)
        print(fobj.getvalue().decode("utf-8").rstrip("\n") + "$")
        fobj.close()

    v = b""
    for n in range(18):
        print()
        v += b"a"
        hexdump(v, ">")

    sys.exit(0)

"""


def trunc(txt, maxlen):
    eoc = azAZ
    ret = u""
    clen = 0
    pend = None
    counting = True
    for input_ofs, ch in enumerate(txt):

        # escape sequences can never contain ESC;
        # treat pend as regular text if so
        if ch == u"\033" and pend:
            clen += len(pend)
            ret += pend
            counting = True
            pend = None

        if not counting:
            ret += ch
            if ch in eoc:
                counting = True
        else:
            if pend:
                pend += ch
                if pend.startswith(u"\033["):
                    counting = False
                else:
                    clen += len(pend)
                    counting = True
                ret += pend
                pend = None
            else:
                if ch == u"\033":
                    pend = unicode(ch)
                else:
                    ret += ch
                    clen += 1

        if clen >= maxlen:
            return [ret, txt[input_ofs:]]

    return [ret, u""]


# adapted from trunc
def strip_ansi(txt):
    eoc = azAZ
    ret = u""
    pend = None
    counting = True
    for ch in txt:

        # escape sequences can never contain ESC;
        # treat pend as regular text if so
        if ch == u"\033" and pend:
            ret += pend
            counting = True
            pend = None

        if not counting:
            if ch in eoc:
                counting = True
        else:
            if pend:
                pend += ch
                if pend.startswith(u"\033["):
                    counting = False
                else:
                    ret += pend
                    counting = True
                pend = None
            else:
                if ch == u"\033":
                    pend = u"{0}".format(ch)
                else:
                    ret += ch
    return ret


# adapted from trunc
def visual_length(txt):
    eoc = azAZ
    clen = 0
    pend = None
    counting = True
    for ch in txt:

        # escape sequences can never contain ESC;
        # treat pend as regular text if so
        if ch == u"\033" and pend:
            clen += len(pend)
            counting = True
            pend = None

        if not counting:
            if ch in eoc:
                counting = True
        else:
            if pend:
                pend += ch
                if pend.startswith(u"\033["):
                    counting = False
                else:
                    clen += len(pend)
                    counting = True
                pend = None
            else:
                if ch == u"\033":
                    pend = unicode(ch)
                else:
                    co = ord(ch)
                    # the safe parts of latin1 and cp437 (no greek stuff)
                    if (
                        co < 0x100  # ascii + lower half of latin1
                        or (co >= 0x2500 and co <= 0x25A0)  # box drawings
                        or (co >= 0x2800 and co <= 0x28FF)  # braille
                    ):
                        clen += 1
                    else:
                        # assume moonrunes or other double-width
                        clen += 2
    return clen


# 83% the speed of visual_length,
# good enough to stop maintaining it and swap w/ len(this)
def visual_indices(txt):
    eoc = azAZ
    ret = []
    pend_txt = None
    pend_ofs = []
    counting = True
    for n, ch in enumerate(txt):

        # escape sequences can never contain ESC;
        # treat pend as regular text if so
        if ch == u"\033" and pend_txt:
            ret.extend(pend_ofs)
            counting = True
            pend_txt = None
            pend_ofs = []

        if not counting:
            if ch in eoc:
                counting = True
        else:
            if pend_txt:
                pend_txt += ch
                pend_ofs.append(n)
                if pend_txt.startswith(u"\033["):
                    counting = False
                else:
                    ret.extend(pend_ofs)
                    counting = True
                pend_txt = None
                pend_ofs = []
            else:
                if ch == u"\033":
                    pend_txt = u"{0}".format(ch)
                    pend_ofs = [n]
                else:
                    ret.append(n)
    return ret


def sanitize_ctl_codes(aside):
    plain = u""
    passthru = (0x02, 0x0B, 0x0F)  # bold, color, reset
    for pch in aside:
        nch = ord(pch)
        # print('read_cb inner  {0} / {1}'.format(b2hex(pch.encode('utf-8', 'backslashreplace')), nch))
        if nch < 0x20 and nch not in passthru:
            print("substituting non-printable \\x{0:02x}".format(nch))
            plain += u"?"
        else:
            plain += pch
    return plain


def sanitize_fn(fn):
    for bad, good in [
        [u"<", u"＜"],
        [u">", u"＞"],
        [u":", u"："],
        [u'"', u"＂"],
        [u"/", u"／"],
        [u"\\", u"＼"],
        [u"|", u"｜"],
        [u"?", u"？"],
        [u"*", u"＊"],
        [u"'", u"＇"],  # shell-safety
        [u"`", u"｀"],  # shell-safety
    ]:
        fn = fn.replace(bad, good)

    if WINDOWS:
        bad = [u"con", u"prn", u"aux", u"nul"]
        for n in range(1, 10):
            bad += u"com{0} lpt{0}".format(n).split(u" ")

        if fn.lower() in bad:
            fn = u"_" + fn

    return fn


FOREGROUNDS = {}
for luma, chars in enumerate([u"01234567", u"89abcdef"]):
    for n, ch in enumerate(chars):
        FOREGROUNDS[ch] = u"\033[{0};3{1}".format(luma, n)

BACKGROUNDS = {}
for n, ch in enumerate(u"01234567"):
    BACKGROUNDS[ch] = u";4{0}".format(n)


def convert_color_codes(txt, preview=False):
    foregrounds = FOREGROUNDS
    backgrounds = BACKGROUNDS
    scan_from = 0
    while txt:
        ofs = txt.find(u"\x0b", scan_from)
        if ofs < 0:
            break

        scan_from = ofs + 1

        fg = None
        if len(txt) > ofs + 1:
            fg = txt[ofs + 1]

        bg = None
        if len(txt) > ofs + 3 and txt[ofs + 2] == u",":
            bg = txt[ofs + 3]

        if fg in foregrounds:
            fg = foregrounds[fg]
        else:
            fg = None
            bg = None  # can't set bg without valid fg

        if bg in backgrounds:
            bg = backgrounds[bg]
        else:
            bg = None

        resume_txt = ofs + 1
        if fg:
            resume_txt += 1
            scan_from = len(fg) + 1
        if bg:
            resume_txt += 2
            scan_from += len(bg)

        preview_k = u""
        if preview:
            resume_txt = ofs + 1
            if fg:
                preview_k = u"K"

        if fg and bg:
            txt = u"%s%s%sm%s%s" % (txt[:ofs], fg, bg, preview_k, txt[resume_txt:])
        elif fg:
            txt = u"%s%sm%s%s" % (txt[:ofs], fg, preview_k, txt[resume_txt:])
        else:
            txt = u"%sK%s" % (txt[:ofs], txt[resume_txt:])

    scan_from = 0
    while txt:
        ofs = txt.find(u"\x02", scan_from)
        if ofs < 0:
            break

        scan_from = ofs + 1
        txt = u"%s\033[1m%s%s" % (
            txt[:ofs],
            u"B" if preview else u"",
            txt[scan_from:],
        )

    scan_from = 0
    while txt:
        ofs = txt.find(u"\x0f", scan_from)
        if ofs < 0:
            break

        scan_from = ofs + 1
        txt = u"%s\033[0m%s%s" % (
            txt[:ofs],
            u"O" if preview else u"",
            txt[scan_from:],
        )

    return txt


"""

def visualize_all_unicode_codepoints_as_utf8():
    stats = [0] * 256
    nmax = sys.maxunicode + 1
    print("collecting all codepoints until {0}d, 0x{1:x}".format(nmax, nmax))

    if PY2:
        to_unicode = unichr  # noqa: F821
        from_char = ord
    else:
        to_unicode = chr
        from_char = int

    for n in range(nmax):
        if n % 0x10000 == 0:
            print(
                "at codepoint {0:6x} of {1:6x},  {2:5.2f}%".format(
                    n, nmax, (100.0 * n) / nmax
                )
            )
        ch = to_unicode(n)

        try:
            bs = ch.encode("utf-8")
        except:
            # python2 allows encoding ud800 as \xed\xa0\x80 which is an illegal sequence in utf8;
            # python -c "for x in unichr(0xd800).encode('utf-8'): print '{0:2x}'.format(ord(x))"
            continue

        for b in bs:
            stats[from_char(b)] += 1

    print()
    for i, n in enumerate(stats):
        v = n
        if v == 0:
            v = "illegal value"
        elif v == 1:
            v = "single-use value"
        print("byte 0x{0:2x} occurences: {1}".format(i, v))
    print()


# visualize_all_unicode_codepoints_as_utf8()

"""


def wrap(txt, maxlen, maxlen2):
    words = txt.rstrip().split()
    ret = []
    for word in words:
        if len(word) * 2 < maxlen or visual_length(word) < maxlen:
            ret.append(word)
        else:
            while visual_length(word) >= maxlen:
                ret.append(word[: maxlen - 1] + u"-")
                word = word[maxlen - 1 :]
            if word:
                ret.append(word)

    words = ret
    ret = []
    ln = u""
    spent = 0
    for word in words:
        wl = visual_length(word)
        if spent + wl > maxlen:
            ret.append(ln[1:])
            maxlen = maxlen2
            spent = 0
            ln = u""
        ln += u" " + word
        spent += wl + 1
    if ln:
        ret.append(ln[1:])

    return ret


def hardwrap(txt, maxlen, vt100):
    # safer than wrap() since it discards escape sequences, making it splittable anywhere
    ret = []
    prefix = u""
    txt = strip_ansi(txt)
    while txt:
        ret.append(prefix + txt[:maxlen])
        txt = txt[maxlen:]
        if vt100:
            prefix = u"\033[36m"

    return ret


def whoops(extra=None):
    msg = r"""
             __
   _      __/ /_  ____  ____  ____  _____
  | | /| / / __ \/ __ \/ __ \/ __ \/ ___/
  | |/ |/ / / / / /_/ / /_/ / /_/ (__  )
  |__/|__/_/ /_/\____/\____/ .___/____/
                          /_/"""
    exc = traceback.format_exc()
    if exc.startswith("None"):
        exc = "".join(traceback.format_stack()[:-1])
    msg = "{0}\r\n{1}\r\n{2}</stack>".format(msg[1:], exc.rstrip(), "-" * 64)
    print(msg)
    if extra:
        print("  {0}\n{1}\n".format(extra, "-" * 64))


def t_a_a_bt():
    ret = []
    for tid, stack in sys._current_frames().items():
        ret.append(u"\r\nThread {0} {1}".format(tid, "=" * 64))
        for fn, lno, func, line in traceback.extract_stack(stack):
            ret.append(u'  File "{0}", line {1}, in {2}'.format(fn, lno, func))
            if line:
                ret.append(u"    {0}".format(line.strip()))
    return u"\r\n".join(ret)


thread_monitor_enabled = False


def monitor_threads():
    global thread_monitor_enabled
    if thread_monitor_enabled:
        return
    thread_monitor_enabled = True

    def stack_collector():
        while True:
            print("capturing stack")
            time.sleep(5)
            txt = t_a_a_bt()
            with open("r0c.stack", "wb") as f:
                f.write(txt.encode("utf-8"))

    thr = threading.Thread(target=stack_collector, name="stk_col")
    thr.daemon = True
    thr.start()


def host_os():
    py_ver = ".".join([str(x) for x in sys.version_info])
    ofs = py_ver.find(".final.")
    if ofs > 0:
        py_ver = py_ver[:ofs]

    bitness = struct.calcsize("P") * 8
    host_os = platform.system()
    return "{0} on {1}{2}".format(py_ver, host_os, bitness)


def py26_threading_event_wait(event):
    """
    threading.Event.wait() is broken on py2.6;
    with multiple subscribers it doesn't always trigger
    """
    if (
        sys.version_info[:2] != (2, 6)
        or INTERP != "CPython"
        or "_Event__flag" not in event.__dict__
    ):
        return

    def nice_meme(timeout=None):
        if event._Event__flag:
            return True

        time.sleep(0.2)
        return event._Event__flag

    event.wait = nice_meme


"""
# ---------------------------------------------------------------------
# dumping ground for mostly useless code below


def test_ansi_annotation():
    rangetype = range
    try:
        rangetype = xrange  # noqa: F405,F821
    except:
        pass
    ansi_txt = (
        "\033[1;33mHello \033[1;32mWorld\033[0m! This \033[7mis\033[0m a test.\033[A"
    )
    ansi_txt = "\033[mf\033[s\033[w\033[has\033[3451431613gt\033[m \033[s\033[g\033[s\033[g\033[s\033[gcod\033[me\033[x"
    rv = visual_indices(ansi_txt)
    print(" ".join(ansi_txt.replace("\033", "*")))
    print(" ".join([str(x % 10) for x in rangetype(len(ansi_txt))]))
    print(" ".join([str(x) for x in rv]))
    print("{0} {1}".format(visual_length(ansi_txt), len(rv)))
    visual = ""
    for ofs in rv:
        visual += ansi_txt[ofs]
    print("[{0}]".format(visual))

    for outer_n in rangetype(3):

        t0 = time.time()
        for n in rangetype(100000):
            rv = visual_indices(ansi_txt)
        print(str(time.time() - t0))

        t0 = time.time()
        for n in rangetype(100000):
            rv = visual_length(ansi_txt)
        print(str(time.time() - t0))
"""



# coding: utf-8
from __future__ import print_function
from .__version__ import S_VERSION, S_BUILD_DT
from .__init__ import EP, PY2, TYPE_CHECKING, unicode
from . import util as Util
from . import chat as Chat
from . import diag as Diag

import re
import time
import base64
import hashlib
import threading
from datetime import datetime

# debug imports
import code
import gc

if TYPE_CHECKING:
    from . import world as World

print = Util.print


HELP_INTRO = u"""\
Useful commands:
   \033[36m/nick\033[0m  change your nickname
   \033[36m/help\033[0m  how-to and about

Text formatting:
  \033[36mCTRL-O\033[0m  reset to default
  \033[36mCTRL-B\033[0m  enable bold/bright
  \033[36mCTRL-K\033[0m  followed by a colour code:
       \033[36m2\033[0m  - \033[32mgreen\033[0m,
    \033[36m15,4\033[0m  - \033[1;37;44mbold white on blue\033[0m --
            say \033[1m/cmap\033[0m to see all options

Switching channels:
  \033[36mCTRL-E\033[0m  jump to active channel
  \033[36mCTRL-A\033[0m  jump to previous channel
  \033[36mCTRL-X\033[0m  jump to next channel
  \033[36m/3\033[0m      go to channel 3
  \033[36m/0\033[0m      go to this channel

Creating or joining the "g" (general) chatroom:
  \033[36m/join #g\033[0m

Leaving a chatroom:
  \033[36m/part\033[0m

Changing your nickname:
  \033[36m/nick new_name\033[0m

Keybinds:
  \033[36mUp\033[0m / \033[36mDown\033[0m       input history
  \033[36mLeft\033[0m / \033[36mRight\033[0m    input field traversing
  \033[36mHome\033[0m / \033[36mEnd\033[0m      input field jump
  \033[36mPgUp\033[0m / \033[36mPgDown\033[0m   chatlog scrolling... \033[1mtry it :-)\033[0m

if you are using a mac, PgUp is fn-Shift-PgUp
if your terminal is tiny, try \033[36m/mn\033[0m and \033[36m/cy\033[0m
"""


class User(object):
    def __init__(self, world, address):
        # type: (World.World, tuple[str, int]) -> User
        self.ar = world.ar
        self.world = world
        self.admin = False  # set true after challenge success
        self.client = None  # the client which this object belongs to
        self.chans = []  # type: list[Chat.UChannel]
        self.active_chan = None  # type: Chat.UChannel
        self.new_active_chan = None  # set for channel change
        self.old_active_chan = None  # last focused channel
        self.nick = None  # type: str
        self.lnick = None  # type: str
        self.nick_re = None  # regex object for ping assert
        self.nick_len = None  # visible segment for self

    def __unicode__(self):
        return u"User {0} {1}".format(self.nick, self.client.adr[0])

    def __str__(self):
        return "User {0} {1}".format(self.nick, self.client.adr[0])

    def __repr__(self):
        return "User({0}, {1})".format(repr(self.nick), repr(self.client.adr[0]))

    def __lt__(self, other):
        return self.nick < other.nick

    def pattern_gen(self, depth=0):
        charset = u"/!@#$%^&*()_+-=[]{};:<>,."
        for ch in charset:
            yield ch

        if depth < 2:  # <= 3 chars
            for ch1 in charset:
                for ch2 in self.pattern_gen(depth + 1):
                    yield ch1 + ch2

    def set_rand_nick(self):
        plain_base = self.ar.nsalt + unicode(self.client.adr[0])
        for suffix in self.pattern_gen():
            plain = plain_base + suffix
            zb = hashlib.sha512(plain.encode("utf-8")).digest()
            zb = base64.b32encode(zb[:20]).lower()
            zs = re.sub("[^a-km-z]", "", zb.decode("utf-8"))
            nv = zs[:6]

            if not self.world.find_user(nv):
                self.set_nick(nv)
                break

    def create_channels(self):
        # while true; do tail -n +3 ansi | iconv -t 'cp437//IGNORE' | iconv -f cp437 | while IFS= read -r x; do printf "$x\n"; done | sed -r "s/$/$(printf '\033[K')/"; printf '\033[J'; sleep 0.2; printf '\033[H'; done

        if self.client.codec in ["utf-8", "cp437", "shift_jis"]:

            # the simple version
            text = u"""\
`1;30m________ ___ ________
`1;30m░▒▓█▀▀▀▀`37m █▀█ `30m▀▀▀▀█▓▒░   `0;36m┌pad1[`0mretr0chat r0c_ver`36m]pad2┐
`1;30m ░▒▓`36m █▀█ █ █ █▀▀ `30m▓▒░    `0;36m│`0mgithub.com/9001/r0c`36m│
`1;30m  ░▒`34m █   █▄█ █▄▄ `30m▒░     `0;36m╘═══════════════════╛
                             `34m  b. r0c_build `0m
"""
            # the messy version
            text = u"""
`1;30m________ `37m__`36m_ `30m________
`1;30m░▒▓█▀▀▀▀`37m █▀`46m▓`0;1;30m ▀▀▀▀█▓▒░   `0;36m┌pad1[`0mret`1mr0c`22mhat r0c_ver`36m]pad2┐
`1;30m ░▒▓ `34;46m▒`0;1;36m▀█ `37;46m▓`0m `1;37;46m▓`0m `1;36m█▀`34m▀ `30m▓▒░    `0;36m│`0mgithub.com/9001/r0c`36m│
`1;30m  ░▒ `34m█   `36m█▄█ `34;46m▒`0;1;34m▄▄ `30m▒░     `0;36m╘═══════════════════╛
                             `34m  b. r0c_build `0m
"""

        else:
            # the simple version
            text = u"""
  `1;37m     /^\\           `0mretr0chat r0c_ver `36m-----
  `1;36m/^^  | |  /^^      `0mgithub.com/9001/r0c
  `1;34m|    \\_/  \\__      `0;36m------b. r0c_build `0m
"""

            # the messy version
            text = u"""
`1;30m______    `37m_`30m    ______
`1;30m\\\\\\\\\\\\\\  `37m/ \\  `30m///////   `0mret`1mr0c`22mhat r0c_ver `36m-----
 `1;30m\\\\ `36m/`37m^^  | |  `36m/^`0;36m^`1;30m //    `0mgithub.com/9001/r0c
  `1;30m\\ `0;36m|    `1m\\_/  `0;36m\\__ `1;30m/     `0;36m------b. r0c_build `0m
"""

        pad1 = pad2 = u"──"
        if len(S_VERSION) > 3:
            pad1 = pad1[1:]
        if len(S_VERSION) > 4:
            pad2 = pad2[1:]
        if len(S_VERSION) > 5:
            pad1 = pad1[1:]
        if len(S_VERSION) > 6:
            pad2 = pad2[1:]

        text = (
            text.replace(u"`", u"\033[")
            .replace(u"r0c_build", S_BUILD_DT)
            .replace(u"r0c_ver", S_VERSION)
            .replace(u"pad1", pad1)
            .replace(u"pad2", pad2)
        )
        text += HELP_INTRO

        # x = u"`1;30m░▒▓█▀▀▀▀`37m█▀`46m▓`0;1;30m▀▀▀▀█▓▒░`0;36m┌ "
        # text += x.replace(u"`", u"\033[") * 66

        uchan = self.world.join_priv_chan(self, u"r0c-status")
        nchan = uchan.nchan
        nchan.topic = u"r0c readme (and status info)"

        msg = Chat.Message(nchan, time.time(), u"-nfo-", text)
        nchan.msgs.append(msg)

        self.new_active_chan = uchan

        self.client.handshake_world = True

    def admin_test(self, cmd, arg):
        if self.admin:
            return True

        print(
            "denied exec:  /{0} {1} from {2} ({3})".format(
                cmd, arg, self.nick, self.client.adr
            )
        )

        inf = self.world.get_priv_chan(self, u"r0c-status").nchan

        self.world.send_chan_msg(
            u"-err-",
            inf,
            u"""\033[1;31m[denied]\033[0m
  don't move, the police are on the way
""",
        )

        return False

    def exec_cmd(self, cmd_str):
        # print('handle {0}'.format(cmd_str))
        inf = self.world.get_priv_chan(self, u"r0c-status").nchan
        cmd = cmd_str  # the command keyword
        arg = None  # single argument with spaces
        arg1 = None  # 1st of 2 arguments
        arg2 = None  # 2nd of 2 arguments

        ofs = cmd.find(u" ")
        if ofs > 0:
            cmd = cmd_str[:ofs]
            arg = cmd_str[ofs + 1 :]
        cmd = cmd.lower()

        if arg:
            arg1 = arg
            ofs = arg.find(u" ")
            if ofs > 0:
                arg1 = arg[:ofs].lower()
                arg2 = arg[ofs + 1 :]

        if cmd == u"help":
            self.help(arg, inf)

        elif cmd == u"me":
            self.world.send_chan_msg(
                u"***",
                self.active_chan.nchan,
                u"\033[1m{0}\033[22m {1}".format(self.nick, arg),
            )

        elif cmd == u"auth":
            if arg == self.ar.pw:
                self.admin = True
                self.world.send_chan_msg(u"-nfo-", inf, u"please don't break anything")
            else:
                self.world.send_chan_msg(u"-err-", inf, u"wrong password")

        elif cmd == u"nick" or cmd == u"n":
            if not arg:
                self.world.send_chan_msg(
                    u"-err-",
                    inf,
                    u"""[invalid argument]
  usage:     /nick  new_nickname
  example:   /nick  spartacus
""",
                )
                return

            legit_chars = Util.azAZ
            legit_chars += u"0123456789"
            legit_within = u"`~!@#$%^&*()_+-=[]{};':\"|,./<>?"
            new_nick = u""
            err_extra = u""
            for ch in arg:
                if ch in legit_chars or (ch in legit_within and new_nick):
                    new_nick += ch
                elif not new_nick:
                    err_extra = u"  (nicks must begin with an alphanumeric character)\n"

            if not new_nick:
                self.world.send_chan_msg(
                    u"-err-",
                    inf,
                    u"[invalid argument]\n  "
                    + u"yooo EXCLUSIVELY illegal chars in new nick\n",
                )
                return

            if new_nick != arg:
                self.world.send_chan_msg(
                    u"-err-",
                    inf,
                    u"[invalid argument]\n  "
                    + u"some illegal characters were removed\n"
                    + err_extra,
                )

            if len(new_nick) > 32:
                self.world.send_chan_msg(
                    u"-err-", inf, u"[invalid argument]\n  " + u"too long\n"
                )
                return

            with self.world.mutex:
                if self.world.find_user(new_nick):
                    self.world.send_chan_msg(
                        u"-err-",
                        inf,
                        u"[invalid argument]\n  " + u"that nick is taken\n",
                    )
                    return

                print(
                    "nick change:  {2} {0} -> {1}".format(
                        self.nick, new_nick, self.client.adr[0]
                    )
                )

                for uchan in self.chans:
                    self.world.send_chan_msg(
                        u"--",
                        uchan.nchan,
                        u"\033[1;36m{0}\033[22m changed nick to \033[1m{1}".format(
                            self.nick, new_nick
                        ),
                        False,
                    )

                # update last-spoke tables
                now = time.time()
                for nchan in [x.nchan for x in self.chans]:
                    nchan.user_act_ts[new_nick] = now
                    try:
                        del nchan.user_act_ts[self.nick]
                    except:
                        pass
                    nchan.update_usernames()

                # update title in DM windows
                for nchan in self.world.priv_ch:
                    for usr in nchan.uchans:
                        if usr.alias == self.nick:
                            usr.alias = new_nick

                self.set_nick(new_nick)

        elif cmd == u"topic" or cmd == u"t":
            if not arg:
                self.world.send_chan_msg(
                    u"-err-",
                    inf,
                    u"""[invalid argument]
  usage:     /topic  the_new_topic
  example:   /topic  cooking recipes
""",
                )
                return

            uchan = self.active_chan
            nchan = uchan.nchan
            if nchan in self.world.priv_ch:
                self.world.send_chan_msg(
                    u"-err-",
                    inf,
                    u"""[error]
  cannot change the topic of private channels
""",
                )
                return

            old_topic = nchan.topic
            nchan.topic = arg
            self.world.send_chan_msg(
                u"--",
                nchan,
                u"\033[36m{0} has changed the topic from [\033[0m{1}\033[36m] -to-> [\033[0m{2}\033[36m]\033[0m".format(
                    self.nick, old_topic, arg
                ),
            )

        elif cmd == u"join" or cmd == u"j":
            if arg and not arg.startswith(u"#"):
                arg = u"#" + arg

            if not arg or len(arg) < 2:
                self.world.send_chan_msg(
                    u"-err-",
                    inf,
                    u"""[invalid arguments]
  usage:     /join  #channel_name
  example:   /join  #general
""",
                )
                return

            nchan = self.world.join_pub_chan(self, arg[1:]).nchan
            # this is in charge of activating the new channel,
            # rapid part/join will crash us without this
            self.client.refresh(False)

        elif cmd == u"part" or cmd == u"p":
            if self.active_chan.alias == u"r0c-status":
                self.world.send_chan_msg(
                    u"-err-",
                    inf,
                    u"""[error]
  cannot part the status channel
""",
                )
                return

            self.world.part_chan(self.active_chan)
            # this is in charge of activating the new channel,
            # rapid part/join will crash us without this
            self.client.refresh(False)

        elif cmd.isdigit():
            nch = int(cmd)
            if nch >= len(self.chans):
                self.world.send_chan_msg(
                    u"-err-",
                    inf,
                    u"""[error]
  you only have {0} channels my dude
""".format(
                        len(self.chans)
                    ),
                )
                return

            self.new_active_chan = self.chans[nch]
            self.client.refresh(False, True)

        elif cmd == u"msg" or cmd == u"m":
            if not arg1 or not arg2:
                self.world.send_chan_msg(
                    u"-err-",
                    inf,
                    u"""[invalid arguments]
  usage:     /msg   nickname   your message text
  example:   /msg   ed   hello world
""",
                )
                return

            if not self.world.find_user(arg1):
                self.world.send_chan_msg(
                    u"-err-",
                    inf,
                    u"""[user not found]
  "{0}" is not online
""".format(
                        arg1
                    ),
                )
                return

            uchan = self.world.join_priv_chan(self, arg1)
            self.new_active_chan = uchan
            self.world.send_chan_msg(self.nick, uchan.nchan, arg2)
            self.client.refresh(False, True)

        elif cmd == u"up" or cmd == u"u":
            self.client.scroll_cmd += -(self.client.h - 4)

        elif cmd == u"down" or cmd == u"d":
            self.client.scroll_cmd += +(self.client.h - 4)

        elif cmd == u"latest" or cmd == u"l":
            self.active_chan.lock_to_bottom = True
            self.client.need_full_redraw = True
            self.client.refresh(False)

        elif cmd == u"view" or cmd == u"v":
            self.client.view = not self.client.view
            self.client.need_full_redraw = True
            self.client.refresh(False)

        elif cmd == u"redraw" or cmd == u"r":
            if self.client.request_terminal_size("redraw"):
                # returns true if event was scheduled for later
                return

            self.client.need_full_redraw = True
            self.client.refresh(False)

        elif cmd == u"cls" or cmd == u"clear":
            if not self.active_chan.alias and not self.admin:
                self.world.send_chan_msg(
                    u"-err-",
                    inf,
                    u"""[error]
  private channels only
""",
                )
                return

            msg = u"\033[33mchannel wiped by \033[1m{0}".format(self.nick)
            if self.active_chan.alias:
                msg = u""

            self.active_chan.nchan.msgs = []
            self.world.send_chan_msg(u"***", self.active_chan.nchan, msg)

        elif cmd == u"fill":
            if not self.admin_test(cmd, arg):
                return

            try:
                nt = int(arg1)
                _ = u"a" in arg2
            except:
                self.world.send_chan_msg(
                    u"-err-",
                    inf,
                    u"""[error]
  bad syntax
""",
                )
                return

            for n in range(int(arg1)):
                self.world.send_chan_msg(
                    self.nick, self.active_chan.nchan, u"{0} {1}".format(arg2, n)
                )

        elif cmd == u"names" or cmd == u"na":
            self.world.send_chan_msg(
                u"--",
                inf,
                u"{1} users in {0}: {2}".format(
                    self.active_chan.nchan.get_name(),
                    len(self.active_chan.nchan.uchans),
                    u", ".join(
                        sorted([x.user.nick for x in self.active_chan.nchan.uchans])
                    ),
                ),
            )

        elif cmd == u"status" or cmd == u"st":
            n_wizard = sum(1 for x in self.world.users if not x.active_chan)
            n_users = len(self.world.users) - n_wizard
            n_pub = len(self.world.pub_ch)
            n_priv = len(self.world.priv_ch) - n_users

            n_in_chans = 0
            seen_users = {}

            for chan in self.world.pub_ch:
                for user in [x.user for x in chan.uchans]:
                    if user not in seen_users:
                        seen_users[user] = 1
                        n_in_chans += 1

            for chan in self.world.priv_ch:
                if len(chan.uchans) == 1:
                    continue
                for user in [x.user for x in chan.uchans]:
                    if user not in seen_users:
                        seen_users[user] = 1
                        n_in_chans += 1

            self.world.send_chan_msg(
                u"--",
                inf,
                u"{0} users + {1} in wizard, {2} in chans, {3} public + {4} private chans".format(
                    n_users, n_wizard, n_in_chans, n_pub, n_priv
                ),
            )

            if self.admin:
                self.world.send_chan_msg(u"--", inf, u"----- users -----")
                for user in sorted(self.world.users):
                    self.world.send_chan_msg(
                        u"--",
                        inf,
                        u"{0} {1} {2}".format(
                            user.client.adr[0].ljust(15),
                            u"ok " if user.active_chan else u"wiz",
                            user.nick,
                        ),
                    )
                self.world.send_chan_msg(u"--", inf, u"----- chans -----")
                for chan in sorted(self.world.pub_ch):
                    self.world.send_chan_msg(
                        u"--",
                        inf,
                        u"{0}: {1}".format(
                            chan.name,
                            u", ".join(sorted([x.user.nick for x in chan.uchans])),
                        ),
                    )
                self.world.send_chan_msg(u"--", inf, u"-----------------")

        elif cmd == u"a":
            activity = {}  # type: dict[int, Chat.UChannel]
            for uchan in self.chans:
                if uchan.hilights and uchan != self.active_chan:
                    activity[uchan.last_ping] = uchan
            for uchan in self.chans:
                if uchan.activity and uchan != self.active_chan:
                    activity[uchan.nchan.msgs[-1].ts] = uchan

            if activity:
                x, uchan = sorted(activity.items())[0]
                self.new_active_chan = uchan
                nchan = uchan.nchan
                for msg in nchan.msgs:
                    if uchan.lock_to_bottom:
                        # skip to end of buffer, users can rely on the
                        # unread-message markers to scroll up if they want
                        break

                    if msg.sno > uchan.last_read:
                        # print('1st unread msg ({0} > {1}) = {2}'.format(
                        # 	msg.sno, uchan.last_read, msg))
                        jump_to = nchan.msgs.index(msg) - 5
                        if jump_to < 0:
                            jump_to = 0
                        uchan.jump_to_msg(jump_to)
                        break
                # print('jumping to activity in {0}'.format(
                # 	self.new_active_chan.nchan.get_name()))

            elif self.old_active_chan:
                ch = self.old_active_chan
                if ch.nchan in self.world.priv_ch or ch.nchan in self.world.pub_ch:
                    self.new_active_chan = ch
                else:
                    self.new_active_chan = self.chans[0]

                # print('jumping to last active, {0}'.format(
                # 	self.new_active_chan.nchan.get_name()))
            else:
                print("cannot jump, no hilights or prev chan")

            self.client.need_full_redraw = True
            self.client.refresh(False)

        elif cmd == u"goto" or cmd == u"g":
            ch = self.active_chan
            nch = self.active_chan.nchan
            if not arg:
                self.world.send_chan_msg(
                    u"--",
                    inf,
                    u"""[goto]
  {1} msgs since {2} in {0}

  command usage:
    /g 19:47             jump to time
    /g 2018-01-21        jump to date
    /g 2018-01-21 19:47  jump to datetime
    /g 3172              jump to message
    /g 34%               jump to offset
    /l                   jump to most recent
""".format(
                        nch.get_name(),
                        len(nch.msgs),
                        nch.msgs[0].dt.strftime("%Y-%m-%d, %H:%M"),
                    ),
                )

            else:
                tfmt = "%Y-%m-%dT%H:%M:%S"

                m = re.match(r"(^[0-9]+)$", arg)
                if m:
                    ch.jump_to_msg(int(m.group(1)))
                    return

                m = re.match(r"(^[0-9\.]+)%$", arg)
                if m:
                    ch.jump_to_msg(int(float(m.group(1)) * len(nch.msgs) / 100.0))
                    return

                m = re.match(r"(^[0-9]{4}-[0-9]{2}-[0-9]{2}) ([0-9]{2}:[0-9]{2})$", arg)
                if m:
                    ht = u"{0}T{1}:00".format(*m.groups())
                    ch.jump_to_time(datetime.strptime(ht, tfmt))
                    return

                m = re.match(r"(^[0-9]{4}-[0-9]{2}-[0-9]{2})$", arg)
                if m:
                    ht = u"{0}T00:00:00".format(m.group(1))
                    ch.jump_to_time(datetime.strptime(ht, tfmt))
                    return

                m = re.match(r"(^[0-9]{2}:[0-9]{2})$", arg)
                if m:
                    ht = u"{0}T{1}:00".format(time.strftime("%Y-%m-%d"), m.group(1))
                    ch.jump_to_time(datetime.strptime(ht, tfmt))
                    return

                self.world.send_chan_msg(
                    u"-err-",
                    inf,
                    u"""[goto]
  invalid argument format, see /g for help
""",
                )

        elif cmd == u"sw":
            try:
                arg = int(arg)
            except:
                arg = None

            if not arg:
                self.world.send_chan_msg(
                    u"-err-",
                    inf,
                    u"""[invalid arguments]
  usage:     /sw  your_screen_width
  example:   /sw  80
""",
                )
                return

            self.client.w = arg
            self.world.send_chan_msg(
                u"-nfo-", inf, u"screen width: {0} letters".format(self.client.w), False
            )

        elif cmd == u"sh":
            try:
                arg = int(arg)
            except:
                arg = None

            if not arg:
                self.world.send_chan_msg(
                    u"-err-",
                    inf,
                    u"""[invalid arguments]
  usage:     /sh  your_screen_height
  example:   /sh  24
""",
                )
                return

            self.client.h = arg
            self.world.send_chan_msg(
                u"-nfo-",
                inf,
                u"screen height: {0} letters".format(self.client.h),
                False,
            )

        elif cmd == u"ss":
            if arg == u"0":
                arg = u"100%"

            try:
                int_arg = int(arg)
            except:
                int_arg = None

            perc_arg = None
            if arg and arg.endswith(u"%"):
                try:
                    perc_arg = int(arg[:-1])
                except:
                    pass

            if int_arg is not None:
                if int_arg > 9000:
                    self.world.send_chan_msg(u"-err-", inf, u"whoa dude")
                    return

                self.client.scroll_f = None
                self.client.scroll_i = int_arg
                self.world.send_chan_msg(
                    u"-nfo-",
                    inf,
                    u"scroll size: {0} lines".format(self.client.scroll_i),
                    False,
                )

            elif perc_arg is not None:
                if perc_arg > 200:
                    self.world.send_chan_msg(u"-err-", inf, u"whoa dude")
                    return

                self.client.scroll_i = None
                self.client.scroll_f = perc_arg / 100.0
                self.world.send_chan_msg(
                    u"-nfo-",
                    inf,
                    u"scroll size: {0}% of screen".format(self.client.scroll_f * 100),
                    False,
                )

            else:
                self.world.send_chan_msg(
                    u"-err-",
                    inf,
                    u"""[invalid arguments]
  usage:     /ss  lines_scrolled_per_pgup_pgdn
  example:   /sh  0     (entire screen)
  example:   /sh  10    (10 lines)
  example:   /sh  50%   (half the screen)
""",
                )
            return

        elif cmd in (u"b?", u"b0", u"b1", u"b2", u"b3"):
            zi = self.client.bell if cmd == u"b?" else int(cmd[1:])
            if zi == 0:
                t = u"disabled. Enable with /b1 or /b2"
            elif zi == 1:
                t = u"enabled for hilights in other channels. Disable with /b0"
            elif zi == 2:
                t = u"enabled for all hilights, even in active channel. Disable with /b0"
            elif zi == 3:
                t = u"enabled for every message, not just hilights. Disable with /b0"

            self.world.send_chan_msg(u"--", inf, u"Audible alerts %s" % (t,), False)
            if self.client.bell != zi:
                self.client.bell = zi
                self.client.save_config()

        elif cmd == u"cy":
            self.client.cnicks = True
            self.client.need_full_redraw = True
            self.world.send_chan_msg(
                u"--", inf, u"Colored nicknames enabled. Disable with /cn", False
            )
            self.client.save_config()

        elif cmd == u"cn":
            self.client.cnicks = False
            self.client.need_full_redraw = True
            self.world.send_chan_msg(
                u"--", inf, u"Colored nicknames disabled. Enable with /cy", False
            )
            self.client.save_config()

        elif cmd == u"my":
            self.client.align = True
            self.client.need_full_redraw = True
            self.world.send_chan_msg(
                u"--", inf, u"Wordwrap margin enabled. Disable with /mn", False
            )
            self.client.save_config()

        elif cmd == u"mn":
            self.client.align = False
            self.client.need_full_redraw = True
            self.world.send_chan_msg(
                u"--", inf, u"Wordwrap margin disabled. Enable with /my", False
            )
            self.client.save_config()

        elif cmd == u"sy":
            try:
                arg = int(arg)
            except:
                arg = 1

            if not self.client.slowmo_tx:
                self.world.cserial += 1

            self.client.slowmo_tx = arg
            m = u"Slowmo enabled ({0}). This avoids a bug in telnet.exe on windows, but your memory is probably busted already so you have to reconnect now to fix it. Disable with /sn if you change your mind"
            self.world.send_chan_msg(u"--", inf, m.format(arg), False)
            self.client.save_config()

        elif cmd == u"sn":
            if self.client.slowmo_tx:
                self.client.slowmo_tx = 0
                self.world.cserial += 1

            self.world.send_chan_msg(
                u"--", inf, u"Slowmo disabled. Enable with /sy", False
            )
            self.client.save_config()

        elif cmd == u"cfg":
            m = []
            keys = "slowmo_tx linemode echo_on vt100 crlf codec bell cnicks align"
            fmt = u"\033[36m{0}=\033[0m{1!r}" if self.client.vt100 else u"{0}={1!r}"
            for k in keys.split(" "):
                m.append(fmt.format(k, getattr(self.client, k)))

            self.world.send_chan_msg(u"--", inf, u" ".join(m), False)

        elif cmd == u"cmap":
            msg = u"All foreground colours (0 to f) on default background,\n"
            msg += u"each code wrapped in [brackets] for readability:\n  "
            for n in range(16):
                if n == 8:
                    msg += u"\n  \033[1;3{0}m[{1:x}], ".format(n % 8, n)
                else:
                    msg += u"\033[3{0}m[{1:x}], ".format(n % 8, n)

            msg += u"\033[0m\n\nEach background with black text:\n  \033[30m"
            for n in range(8):
                msg += u"\033[4{0}m 0,{0} ".format(n)

            msg += u"\033[0m\n\nEach background with gray text:\n  \033[37m"
            for n in range(8):
                msg += u"\033[4{0}m 7,{0} ".format(n)

            msg += u"\033[0m\n\nEach background with white text:\n  \033[1;37m"
            for n in range(8):
                msg += u"\033[4{0}m f,{0} ".format(n)

            msg += u"\033[0m\n"
            self.world.send_chan_msg(u"-nfo-", inf, msg)

        elif cmd == u"bc":
            if not self.admin_test(cmd, arg):
                return

            self.world.broadcast_message(u"\033[31m{0}".format(arg), 2)

        elif cmd == u"sd":
            if not self.admin_test(cmd, arg):
                return

            msg = u"\033[31mserver shutdown requested by \033[1m{0}".format(self.nick)
            self.world.broadcast_message(msg, 2)

            def delayed_shutdown():
                time.sleep(0.5)
                self.world.core.shutdown()

            thr = threading.Thread(target=delayed_shutdown, name="shutd")
            thr.daemon = True
            thr.start()

        elif cmd == u"mem":
            if not self.admin_test(cmd, arg):
                return

            print("memdump started")
            Diag.memory_dump()
            print("memdump done")

        elif cmd == u"repl":
            if not self.admin_test(cmd, arg):
                return

            print("entering repl")
            # code.interact(locals=locals())
            code.InteractiveConsole(locals=globals()).interact()
            print("left repl")

        elif cmd == u"gc":
            if not self.admin_test(cmd, arg):
                return

            gc.collect()

        elif cmd == u"quit" or cmd == u"q" or cmd == u"exit":
            self.client.host.part(self.client)

        else:
            self.world.send_chan_msg(
                u"-err-",
                inf,
                u"""invalid command:  /{0}
  if you meant to send that as a message,
  escape the leading "/" by adding another "/"
""".format(
                    cmd_str
                ),
            )

    def set_nick(self, new_nick):
        if self.lnick in self.world.lusers:
            del self.world.lusers[self.lnick]

        nick_re = u""
        # re.IGNORECASE doesn't work
        # this is dumb
        for ch in re.escape(new_nick):
            if ch not in Util.azAZ:
                nick_re += ch
            else:
                nick_re += u"[{0}{1}]".format(ch.lower(), ch.upper())
                # surprisingly equally fast or faster than haystack.lower()

        self.nick = new_nick
        self.lnick = new_nick.lower()
        self.nick_re = re.compile("(^|[^a-zA-Z0-9]){0}([^a-zA-Z0-9]|$)".format(nick_re))

        self.nick_len = len(new_nick)
        if self.nick_len > self.client.w * 0.25:
            self.nick_len = int(self.client.w * 0.25)

        self.world.lusers[self.lnick] = self

        if self.active_chan:
            self.client.save_config()

    def help(self, arg, inf):
        if not arg:
            arg = u"topics"

        txt = None
        if arg == u"intro":
            txt = HELP_INTRO
        else:
            legit_chars = Util.azAZ
            page = u""
            for ch in arg:
                if ch in legit_chars:
                    page += ch
            try:
                with open(EP.doc + "help-{0}.md".format(page), "rb") as f:
                    txt = f.read().decode("utf-8")
            except:
                self.world.send_chan_msg(
                    u"-err-", inf, u"that help page does not exist"
                )
                return

        txt = txt.replace(u"\r", u"")
        txt = txt.replace(u"  \n", u"\n")

        txt = u"\033[0;30;46m{0}\033[K\n\033[0m\n".format(u"=" * 32) + txt

        txt = txt.replace(u"\n\n| | |\n|-|-|\n", u"\n")
        if u"\n| " in txt:
            txt = txt.replace(u"\n| ", u"\n").replace(u" | ", u"  ")
            if txt.startswith(u"| "):
                txt = txt[2:]

        txt = re.sub(r"\*\*`?([^\*`]+)`?\*\*", "\033[1;36m\\1\033[0m", txt)
        txt = re.sub(r"`([^`]+)`", "\033[1;35m\\1\033[0m", txt)
        txt = re.sub(r"\n# ([^\n]*)\n", "\n\033[1;33m=== \\1 ===\033[0m\n", txt)
        txt = txt.replace(u"\n", u"\r\n")

        self.world.send_chan_msg(u"-nfo-", inf, txt)



# coding: utf-8
from __future__ import print_function
from .__init__ import TYPE_CHECKING
from . import util as Util

from datetime import datetime
import calendar
import operator

if TYPE_CHECKING:
    from . import ivt100 as Ivt100
    from . import user as User

print = Util.print


class NChannel(object):
    def __init__(self, name, topic):
        # type: (str, str) -> NChannel
        self.uchans = []  # type: list[UChannel]
        self.msgs = []  # type: list[Message]
        self.name = name
        self.topic = topic
        self.topic_bak = None
        self.user_act_ts = {}  # type: dict[str, int]  # str(nick) -> ts(last activity)
        self.usernames = u""  # sorted by activity

        self.log_fh = None  # active log file
        self.log_ctr = 0  # number of messages in file

    def get_name(self):
        if self.name:
            return u"#" + self.name
        ret = u", ".join(x.alias for x in self.uchans[:2])
        return ret or u"<abandoned private channel>"

    def __unicode__(self):
        return u"NChannel {0}".format(self.get_name())

    def __str__(self):
        return "NChannel {0}".format(self.get_name())

    def __repr__(self):
        return "NChannel({0}, {1})".format(self.name, repr(self.topic))

    def __lt__(self, other):
        return self.name < other.name

    def update_usernames(self):
        self.usernames = u", ".join(
            [
                k
                for k, _ in sorted(
                    self.user_act_ts.items(),
                    key=operator.itemgetter(1),
                    reverse=True,
                )[:32]
            ]
        )


class UChannel(object):
    def __init__(self, user, nchan, alias=None):
        # type: (Ivt100.VT100_Client, User.User, NChannel, str) -> UChannel
        self.user = user  # the user which this object belongs to
        self.nchan = nchan  # the NChannel object
        self.alias = alias  # local channel name (private)
        self.last_read = 0  # most recent sno viewed in this channel
        self.last_ping = 0  # most recent sno that was a hilight
        self.hilights = False
        self.activity = False
        self.display_notification = False
        self.lock_to_bottom = True
        self.vis = []  # type: list[VisMessage]  # visible messages

    def update_activity_flags(self, set_last_read=False, last_nchan_msg=0):
        if not last_nchan_msg and self.nchan.msgs:
            last_nchan_msg = self.nchan.msgs[-1].sno

        if set_last_read:
            if self.vis:
                self.last_read = max(self.last_read, self.vis[-1].msg.sno)
            else:
                self.last_read = last_nchan_msg

        hilights = self.last_read < self.last_ping
        activity = self.last_read < last_nchan_msg

        self.display_notification = hilights or activity

        if (
            self.display_notification
            and self == self.user.active_chan
            and self.lock_to_bottom
        ):
            # don't display notifications in the status bar
            # when chan is active and bottom messages are visible
            self.display_notification = False

        if hilights:
            self.user.client.notify_new_hilight(self)
        elif activity:
            self.user.client.beep(3)

        if self.hilights != hilights or self.activity != activity:
            self.hilights = hilights
            self.activity = activity
            return True

        return False

    def jump_to_msg(self, msg_n):
        if msg_n >= len(self.nchan.msgs):
            msg_n = len(self.nchan.msgs) - 1

        self.vis = [
            VisMessage().c_new(self.nchan.msgs[msg_n], [u"x"], msg_n, 0, 1, self)
        ]

        self.lock_to_bottom = False
        self.user.client.need_full_redraw = True

    def jump_to_time(self, dt):
        ts = calendar.timegm(dt.timetuple())
        for msg in self.nchan.msgs:
            if msg.ts >= ts:
                i = self.nchan.msgs.index(msg)
                print("jump to {0} of {1}".format(i, len(self.nchan.msgs)))
                return self.jump_to_msg(i)

        return self.jump_to_msg(len(self.nchan.msgs) - 1)


class VisMessage(object):
    def __init__(self):
        pass

    def c_new(self, msg, txt, im, car, cdr, ch):
        # type: (Message, list[str], int, int, int, UChannel) -> VisMessage
        self.msg = msg  # the message object
        self.txt = txt  # the formatted text
        self.im = im  # offset into the channel's message list
        self.car = car  # first visible line
        self.cdr = cdr  # last visible line PLUS ONE
        self.cli = ch.user.client

        if not msg or not msg.user:
            Util.whoops("msg bad")
        if not ch or not ch.user:
            Util.whoops("user bad")

        self.unformatted = txt[0]
        self.hilight = bool(ch.user.nick_re.search(msg.txt))
        self.unread = msg.user != ch.user.nick and msg.sno > ch.last_read

        # print('add msg for {0} which is unread {1}, hilight {2}'.format(
        # 	ch.user.nick, self.unread, self.hilight))

        self.apply_markup()
        return self

    def c_segm(self, other, src_car, src_cdr, new_car, new_cdr, ch):
        self.msg = other.msg
        self.txt = other.txt[src_car:src_cdr]
        self.im = other.im
        self.car = new_car
        self.cdr = new_cdr
        self.cli = ch.user.client

        self.hilight = other.hilight
        self.unread = other.unread
        if src_car == 0:
            self.unformatted = other.unformatted
        else:
            self.unformatted = self.txt[0]

        return self

    def plaintext(self):
        return [self.unformatted] + self.txt[1:]

    def apply_markup(self):
        ln = self.unformatted
        if self.cli.vt100:
            if self.hilight and self.unread:
                prefix = u"\033[1;35;7m"
                postfix = u"\033[0m "
            elif self.hilight:
                prefix = u"\033[1;35m"
                postfix = u"\033[0m "
            elif self.unread:
                prefix = u"\033[7m"
                postfix = u"\033[0m "
            else:
                prefix = None
                postfix = None
        else:
            prefix = u""
            if self.hilight and not self.cli.view:
                postfix = u"="
            else:
                postfix = None

        if self.cli.view:
            if postfix:
                self.txt[0] = u"%s%s%s" % (prefix, ln, postfix)
            elif self.cli.vt100 and u"\033" not in ln:
                ofs = 1 + len(ln) - len(ln.lstrip())
                self.txt[0] = u"\033[0;1m%s\033[0;36m%s\033[0m" % (ln[:ofs], ln[ofs:])
            else:
                self.txt[0] = ln
        elif postfix and not ln.startswith(u" "):
            ofs = ln.find(u" ")
            self.txt[0] = u"%s%s%s%s" % (prefix, ln[:ofs], postfix, ln[ofs + 1 :])
        else:
            self.txt[0] = ln


class Message(object):
    def __init__(self, to, ts, user, txt):
        # type: (NChannel, int, str, str) -> Message
        self.ts = ts  # int timestamp; 1M msgs = 38MiB
        self.dt = datetime.utcfromtimestamp(ts)  # 1M msgs = 53MiB
        self.user = user  # str username
        self.txt = txt  # str text

        # set serial number based on last message in target
        if to and to.msgs:
            self.sno = to.msgs[-1].sno + 1
        else:
            self.sno = 0

    def __unicode__(self):
        hhmmss = self.dt.strftime("%Y-%m%d-%H%M%S")
        return u"Message {0:x} time({1},{2}) from({3}) body({4})".format(
            id(self), self.ts, hhmmss, self.user, self.txt
        )

    def __str__(self):
        hhmmss = self.dt.strftime("%Y-%m%d-%H%M%S")
        return "Message {0:x} time({1},{2}) from({3}) body({4})".format(
            id(self), self.ts, hhmmss, self.user, self.txt
        )

    def __repr__(self):
        return "Message({0}, '{1}', {2})".format(self.ts, self.user, repr(self.txt))
